program Ultimatch_new_all
  
    use sep
    use cartesian
    use row_findnsmatch_mod
    use row_findmatch_mod
    use array_findmatch_mod
    use array_findnsmatch_mod
    use array_ncnhconest_mod
    
    use helix
    use nhelix
    use nhelicon
    use nchelicon
    use nchconest
    use lapfac
    use findmatch_mod
    use verbosity_mod

    implicit none

    real,pointer, dimension(:)   :: muls,prims,target,out1,out2,weight,dummy_muls,input
    type (filter)                :: match1,match2,laplac,laplac1,laplac2
    type (nfilter)               :: nsmatch1,nsmatch2

    !! input parameters
    integer :: ndim              ! dimension of dataset
    integer :: mode              ! 1=single filter, 2=NS filtering
    integer :: wmode             ! 1= no weight 2=L1 3=user-provided weight
    integer :: nsmode            ! 1=regularized 2=precond 3=no-reg-no-precond
    integer,dimension(:),pointer  :: nitr ! number of iterations (size nouter)
    integer :: op                ! =1 normal operation 2=row operator 3=array operator 
    real,dimension(:),pointer :: mu1 ! balance fitting of prims and mults (nouter) 
    real,dimension(:),pointer :: epss ! for the regularization term (nouter) 
	real :: eps,mu
	integer :: niter
    real    :: eps1,eps2,eps3    ! to be used only if op=3 to balance the three goals
    logical :: twodim
    logical :: dottest
    integer, dimension(:), pointer :: npatch ! no. of patches, dimension(ndim)

    integer, dimension(:), pointer :: nfilt  ! no. of coefficients in each filter
    integer, dimension(:), pointer :: n      ! data-size, dimension(ndim)
    real,    dimension(:), pointer :: d,o
    integer, dimension(:), pointer :: nwind  ! overlapping patch-size
    integer, dimension(:), pointer :: idx    ! index counter (ndim)

    integer, dimension(:), pointer :: nh     ! length of each filter in the table
    integer, dimension(:), pointer :: pch    ! ns filter indexing
    integer, dimension(:), pointer :: psize  ! non-overlapping patch-size

    integer    :: i,stat,ilag,ipatch,idim,ipanels,npanels ! junk variables
    integer    :: nhmax
    real       :: min_mask,max_mask
    integer    :: j,jj,ii,nouter
	logical    :: switch
  
    call sep_init(SOURCE)

    ndim=sep_dimension()
write(0,*)'NDIM=',ndim
    allocate(n(ndim), o(ndim), d(ndim))
    
    !! parameters from history file (target)
    call from_history('n',n)
    call from_history('o',o)
    call from_history('d',d)

    !! parameters to output history file (multiples)
    call to_history('n',n)
    call to_history('o',o)
    call to_history('d',d)
write(0,*)'n=',n,'o=',o,'d=',d

    !! read matching parameters from the command line and set defaults
    call from_param('mode',   mode,   1     )
    call from_param('wmode',  wmode,  1     )
    call from_param('nsmode', nsmode, 1     )
    call from_param('2D',     twodim, .false.)
    call from_param('dtest',  dottest,.false.)
    call from_param('verb',   verb,   10    )
    call from_param('op',     op,     1     )
    call from_param('eps1',   eps1,   1.    ) ! for the array operator
    call from_param('eps2',   eps2,   1.    ) ! for the array operator
    call from_param('eps3',   eps3,   1.    ) ! for the array operator
    call from_param('nouter',nouter,  1     )
    call from_param('switch',switch, .false.     )
	allocate(epss(nouter),nitr(nouter),mu1(nouter))
	if (nouter >1) then
    	call from_param('nitr',  nitr       ) ! number of linear iterations
    	call from_param('mu1',  mu1         ) 
    	call from_param('epss',  epss       ) 
	else if (nouter==1) then
    	call from_param('mu',     mu,     1.) ! for the row operator
		mu1(1)=mu
    	call from_param('eps',    eps,    0.) ! for the regularization term
		epss(1)=eps
	else
		call seperr("nouter cannot be negative")
	end if
		
    if (wmode==3 .AND. op>1) call seperr ("ERROR: can't use wmode=3 and op>1")
    if (wmode==2) write(0,*)'WARNING: *****************************'
    if (wmode==2) write(0,*)'WARNING: wmode=2 is an untested option'
    if (wmode==2) write(0,*)'WARNING: *****************************'
    if (op==1) mu=1.

write(0,*)'2D=',twodim,'verb=',verb,'op=',op,'mu=',mu,'eps=',eps

    !! if primaries requested as output, write out the history file parameters
    if (op >1 .AND. exist_file("primaries_out")) then
        call to_history('n1',n(1),"primaries_out")
        call to_history('d1',d(1),"primaries_out")
        call to_history('o1',o(1),"primaries_out")
        if (ndim>1) then
            call to_history('n2',n(2),"primaries_out")
            call to_history('d2',d(2),"primaries_out")
            call to_history('o2',o(2),"primaries_out")
        end if
        if (ndim>2) then
            call to_history('n3',n(3),"primaries_out")
            call to_history('d3',d(3),"primaries_out")
            call to_history('o3',o(3),"primaries_out")
        end if
    end if

    !! check number of dimensions and initialize correspondingly
    npanels=1
    if (twodim) then
        allocate(nfilt(2))
        nfilt=n(1:2)
write(0,*)'NFILT=',nfilt
        npanels=1
        if (ndim>2) npanels=n(3)
        deallocate(n)
        ndim=2
write(0,*)'NPANELS=',npanels
        allocate(n(2))
        n=nfilt
        deallocate(nfilt)
    end if

    allocate(nfilt(ndim),idx(ndim))
    nfilt=1; nfilt(1)=21                !! default fitler length

    call from_param('nfilt',nfilt,nfilt)
write(0,*)'nfilt=',nfilt
	if (nouter==1) then
		call from_param('niter',niter,product(nfilt)-1)
		nitr(1)=niter
	end if

    !! Patching parameters
    if (mode /= 1) then
        allocate(npatch(ndim),nwind(ndim))
        npatch=1
        call from_param('npatch',npatch,npatch)
    end if

    ! allocate space for the multiples and optionally the primaries
    allocate(muls(product(n)))
    if (op>1) allocate(prims(product(n)))
    allocate(out1(product(n)),out2(product(n)),input(product(n)))
    if (op==3) then         !! array operator
        allocate(target(3*product(n)))
    else
        allocate(target(product(n)))
    end if     

    ! Create TWO filters, one to match primaries and one to match multiples 
    call allocatehelix(match1,product(nfilt))
    if (op>1) call allocatehelix(match2,product(nfilt))
    do ilag=1,product(nfilt)
        call line2cart(nfilt,ilag,idx)
        match1%lag(ilag)=idx(1)-nfilt(1)/2-1
        if (op>1) match2%lag(ilag)=idx(1)-nfilt(1)/2-1
        do idim=2,ndim
            match1%lag(ilag)=match1%lag(ilag)+(idx(idim)-nfilt(idim)/2-1)&
                & *product(n(1:idim-1))
            if (op>1) &
                match2%lag(ilag)=match2%lag(ilag)+(idx(idim)-nfilt(idim)/2-1)&
                & *product(n(1:idim-1))
        end do
    end do

    select case (mode)
    
    case (1)
        if (verb>0) write(0,*) '############################'
        if (verb>0) write(0,*) '# Single matched-filter...  '
        if (verb>0) write(0,*) '#---------------------------'
        if (verb>1) write(0,*) '#    nfilt  = ',nfilt
        if (verb>0) write(0,*) '############################'

    case (2)    
        if (verb>0) write(0,*) '#####################################'
        if (verb>0) write(0,*) '# Non-stationary matched-filtering...'
        if (verb>0) write(0,*) '#------------------------------------'
        if (verb>1) write(0,*) '#    ndim   = ',ndim
        if (verb>1) write(0,*) '#    n      = ',n
        if (verb>1) write(0,*) '#    npatch = ',npatch
        if (verb>1) write(0,*) '#    nfilt  = ',nfilt

        !! Create a non-stationary filter
        allocate(nh(product(npatch))) !! nh: length of each filter in table
        nh=product(nfilt)             !! Set them all constant
        call psize_init(n,npatch,psize) !! psize(ndim) is the patch-size
     
        if (verb>0) write(0,*) '#    patch-size = ',psize
        if (verb>0) write(0,*) '#####################################'
     
        call pch_init(n,npatch,psize,pch)  !! pch(ndata) is the index
        call nallocate(nsmatch1,nh,pch)
        if (op>1) call nallocate(nsmatch2,nh,pch)

        !! initialize the non-stationary filters
        do ipatch=1,product(npatch)
            nsmatch1%hlx(ipatch)%lag=match1%lag
            if (op>1) nsmatch2%hlx(ipatch)%lag=match2%lag
        end do

        !! compute the helix version of the laplacian regularizer
        nhmax=size(nsmatch1%hlx(1)%lag)
        if (op>1) then
            nhmax=nhmax+size(nsmatch2%hlx(1)%lag)
            call laplac1_init(laplac,nhmax,npatch) !! laplac is the regularizer
        else
            call laplac_init(laplac,nsmatch1,npatch) !! laplac is the regularizer
        end if

    case default
        call seperr('mode not implemented')
    end select    

    !! compute the filters panel by panel
    panels:do ipanels=1,npanels

        if (ipanels >1) dottest=.false.

        if (verb>1) write(0,*) '###################################'
        if (verb>1) write(0,*) '   Panel = ',ipanels,' out of ',npanels
        if (verb>1) write(0,*) '-----------------------------------'

        out1=0.
        out2=0.
     
        ! Read in data
        call sep_read(muls,'multiples_in')
        if (op>1) call sep_read(prims,'primaries_in')
        call sep_read(input)
        if (op==3) then
            target(1:product(n))=input
            target(product(n)+1:2*product(n))=input
            target(2*product(n)+1:3*product(n))=input
        else
            target=input
        end if

        !! if the user supplied a weight (mask), make sure that it makes
        !! sense and scale by its maximum value
        if (wmode==3) then
            allocate(weight(product(n)))
            call sep_read(weight,'weight')
            max_mask=-9999; min_mask=99999
            do ii=1,product(n)
                if (weight(ii)>max_mask) max_mask=weight(ii)
                if (weight(ii)<max_mask) min_mask=weight(ii)
            end do
            if (max_mask==0. .OR. min_mask<0) then
                write(0,*)'ERROR:mask is all zeros or contain negative numbers'
                stop
            end if
write(0,*)'max_mask=',max_mask
write(0,*)'min_mask=',min_mask
            weight=weight/max_mask
        end if

        select case(mode)

        case (1) !! one filter

            select case(wmode)
                    
            case (1) !! no weight
                if (verb>2) write(0,*) ' -----------------------------------'
                if (verb>2) write(0,*) ' Calling findmatch with no weight...'
                if (verb>2) write(0,*) ' -----------------------------------'

                select case(op)

                case (1) !! standard operation. Match multiples only
                    if (verb>2) write(0,*) ' -----------------------------------'
                    if (verb>2) write(0,*) ' Matching multiples only ...'
                    if (verb>2) write(0,*) ' -----------------------------------'

                    call findmatch(muls,target,match1,niter,dottest)
                    call applymatch(muls,out1,match1,dottest)
                    call dumpfilter(match1)

                case (2) !! match multiples and primaries wih row operator

                    if (verb>2) write(0,*) ' -----------------------------------'
                    if (verb>2) write(0,*) ' Matching multiples and primaries '
                    if (verb>2) write(0,*) ' with row operator [muM P][m1'
                    if (verb>2) write(0,*) '                           m2]=d'
                    if (verb>2) write(0,*) ' -----------------------------------'

                    call row_findmatch(muls,prims,target,match1,match2,&
                        & niter,mu,dottest)
                    call dumpfilter(match1)
                    call dumpfilter(match2)

                    !! apply the first filter to the data to get 
                    !!the matched multiples
                    call applymatch(muls,out1,match1,dottest)

                    !! if requested, apply the second filter to the data to get 
                    !!    the matched primaries. No dottest
                    if (exist_file("primaries_out")) then
                        call applymatch(prims,out2,match2,.false.) 
                    end if

                case (3) !! match multiples and primaries with array operator
                    if (verb>2) write(0,*) ' -----------------------------------'
                    if (verb>2) write(0,*) ' Matching multiples and primaries '
                    if (verb>2) write(0,*) ' with array operator [muM P]   =[d1]'
                    if (verb>2) write(0,*) '                     [muM 0][m]=[d2]'
                    if (verb>2) write(0,*) '                     [0   P]   =[d3]'
                    if (verb>2) write(0,*) ' -----------------------------------'

                    call array_findmatch(muls,prims,target,match1,match2,&
                        & niter,mu,eps1,eps2,eps3,dottest)

                    call dumpfilter(match1)
                    call dumpfilter(match2)

                    !! apply the first filter to the data to get 
                    !!the matched multiples
                    call applymatch(muls,out1,match1,dottest)
                    out1=eps2*out1

                    !! if requested, apply the second filter to the data to get 
                    !!    the matched primaries. No dottest
                    if (exist_file("primaries_out")) then
                        call applymatch(prims,out2,match2,.false.) 
                        out2=eps3*out2 
                    end if

                case default
                    call seperr('Operator not implemented. op has to be 1,2,3')
                end select
            
            case (3) !! user-supplied weight

                if (verb>2) write(0,*)' -----------------------------------'
                if (verb>2) write(0,*)' Calling findmatch with user-supplied weight...'
                if (verb>2) write(0,*)' -----------------------------------'

                !! outer cycle (non-linear iterations)
                do jj=1,nouter
                    call findmatch(muls,target,match1,niter,dottest,wght=weight)
                    call dumpfilter(match1)
                    if (nouter>1) then
                        dummy_muls=muls
                        call nchconest_init(dummy_muls,match1)
                        stat=nchconest_lop(.FALSE.,.FALSE.,dummy_muls,match1%flt)
                        weight=dummy_muls-muls
                        weight=abs(1.-weight)
                    end if
                end do
                call applymatch(muls,out1,match1,dottest)

            case default
                call seperr('Weight mode not implemented')
            end select

            !! write out the filter and the multiples
            if (wmode==3) out1=out1*weight !! apply the mask to the multiples

        case (2) !! NS filters
           
            select case (wmode)

            case (1)    !! no weight
            
                if (verb>2) write(0,*) ' -------------------------------------'
                if (verb>2) write(0,*) ' Calling findnsmatch with no weight...'
                if (verb>2) write(0,*) ' -------------------------------------'
           
                select case(op)    

                case (1) !! standard operation. Match multiples only
                    if (verb>2) write(0,*) ' -----------------------------------'
                    if (verb>2) write(0,*) ' Matching multiples only ...'
                    if (verb>2) write(0,*) ' -----------------------------------'

                    call findnsmatch(muls,target,nsmatch1,laplac,nsmode,&
                        & niter,eps,dottest)    
                    call applynsmatch(muls,out1,nsmatch1,dottest) 

                case (2) !! match multiples and primaries wih row operator

                    if (verb>2) write(0,*) ' -----------------------------------'
                    if (verb>2) write(0,*) ' Matching multiples and primaries '
                    if (verb>2) write(0,*) ' with row operator [muM P][m1'
                    if (verb>2) write(0,*) '                           m2]=d'
                    if (verb>2) write(0,*) ' -----------------------------------'

                    !! the 1 is for regularized solver.
                    call row_findnsmatch(muls,prims,target,nsmatch1,nsmatch2,&
                        & laplac,1,nitr,epss,mu1,nouter,dottest,n,o,d,switch)
           
                    !! apply the first filter to the data to get the matched multiples
                    call applynsmatch(muls,out1,nsmatch1,dottest) 
        
                    !! if requested, apply the second filter to the data to get 
                    !!    the matched primaries
                    if (exist_file("primaries_out")) then
                        call applynsmatch(prims,out2,nsmatch2,.false.) 
                    end if

                case (3) !! match multiples and primaries with array operator

                    if (verb>2) write(0,*) ' -----------------------------------'
                    if (verb>2) write(0,*) ' Matching multiples and primaries '
                    if (verb>2) write(0,*) ' with array operator [muM P]   =[d1]'
                    if (verb>2) write(0,*) '                     [muM 0][m]=[d2]'
                    if (verb>2) write(0,*) '                     [0   P]   =[d3]'
                    if (verb>2) write(0,*) ' -----------------------------------'

                    !! the 1 is for regularized solver.
                    call array_findnsmatch(muls,prims,target,nsmatch1,nsmatch2,&
                        & laplac,1,niter,eps,mu,eps1,eps2,eps3,dottest)
           
                    !! apply the first filter to the data to get the matched multiples
                    call applynsmatch(muls,out1,nsmatch1,dottest) 
                    out1=eps2*out1
        
                    !! if requested, apply the second filter to the data to get 
                    !!    the matched primaries
                    if (exist_file("primaries_out")) then
                        call applynsmatch(prims,out2,nsmatch2,.false.) 
                        out2=eps3*out2
                    end if

                case default
                    call seperr('Operator not implemented. op has to be 1,2,3')
                end select
            
            case (3) ! user-supplied weight

                if (verb>2) write(0,*) ' -----------------------------'
                if (verb>2) write(0,*) ' Calling findnsmatch with user-supplied weight...'
                if (verb>2) write(0,*) ' -----------------------------'

                call findnsmatch(muls,target,nsmatch1,laplac,nsmode,niter,&
                    & eps,dottest,wght=weight)
                call applynsmatch(muls,out1,nsmatch1,dottest) 

            case default
                call seperr('Weight mode not implemented')
            end select

        case default
            call seperr('mode not implemented')
        end select

        out1=mu*out1
        call sep_write(out1)
            
        if (op>1) then
            if (exist_file("primaries_out")) then
                call srite("primaries_out",out2,4*size(out2))
            end if
        end if
    end do panels
  
    write(0,*) '---- The End ----'
  
    call sep_close()
  
    contains
     
    !! Calculate non-overlapping patch-size given a certain number of patches
    subroutine psize_init(n,npatch,psize)
        integer, dimension(:), intent(in)  :: n,npatch
        integer, dimension(:), pointer     :: psize
        allocate(psize(ndim))
        do i=1,ndim
            psize(i)=1+(n(i)-1)/npatch(i)
        end do
    end subroutine psize_init

    !! Divides data-space between rectangular patches
    subroutine pch_init(n,npatch,psize,pch)
        integer, dimension(:), intent(in)  :: n,npatch,psize
        integer, dimension(:), pointer     :: pch,ii
        integer :: i,j

        allocate(pch(product(n)),ii(size(n)))
        do i=1,product(n)
            pch(i)=0
            call line2cart(n,i,ii)
            do j=1,size(n)
                ii(j)=(1+(ii(j)-1)/psize(j))
            end do
            call cart2line(npatch,ii,pch(i))
        end do
        deallocate(ii)
    end subroutine pch_init
  
    !! Create a roughener to dampen NS filter estimation
    subroutine laplac_init(laplac,nsfilt,npatch)
        type  (filter)        :: laplac
        type (nfilter)        :: nsfilt
        integer, dimension(:) :: npatch
        real    :: e1,e2
        integer :: nhmax

        nhmax=size(nsfilt%hlx(1)%lag)

        write(0,*) 'size',size(npatch)
        select case(size(npatch))
            case (1)
                call allocatehelix(laplac,1)  
                laplac%lag(1)=nhmax;   laplac%flt(1)=-.99
            case (2)
                !! 1-D smoothing only

                ! crap 2-D smoothing
                call allocatehelix(laplac,3)  
                laplac%lag=(/ nhmax, npatch(1)*nhmax, nhmax*(npatch(1)+1)  /)
                e1=0.2; e2=0.1
                laplac%flt=(/ e1-1.,      e2-1.,     (1.-e1)*(1.-e2)  /)

                if (verb>5) then
                    write(0,*) '  Roughening filter...'
                    call dumpfilter(laplac)
                end if

            case default
                write(0,*) 'Testmatch'
                call seperr('dimensionality not implemented')
        end select

    end subroutine laplac_init

    !! Create a roughener to dampen NS filter estimation
    subroutine laplac1_init(laplac,nhmax,npatch)
        type  (filter)            :: laplac
        integer                    :: nhmax
        integer, dimension(:)    :: npatch
        real                    :: e1,e2

        write(0,*) 'size',size(npatch)
        select case(size(npatch))
            case (1)
                call allocatehelix(laplac,1)  
                laplac%lag(1)=nhmax;   laplac%flt(1)=-.99
            case (2)
                !! 1-D smoothing only

                ! crap 2-D smoothing
                call allocatehelix(laplac,3)  
                laplac%lag=(/ nhmax, npatch(1)*nhmax, nhmax*(npatch(1)+1)  /)
                e1=0.2; e2=0.1
                laplac%flt=(/ e1-1.,      e2-1.,     (1.-e1)*(1.-e2)  /)

                if (verb>5) then
                    write(0,*) '  Roughening filter...'
                    call dumpfilter(laplac)
                end if

            case default
                write(0,*) 'Testmatch'
                call seperr('dimensionality not implemented')
            end select

        end subroutine laplac1_init

        subroutine row_laplac_init (laplac1,laplac2,nsmatch1,nsmatch2,npatch)
            type  (filter)        :: laplac1
            type  (filter)        :: laplac2
            type (nfilter)        :: nsmatch1
            type (nfilter)        :: nsmatch2
            integer, dimension(:) :: npatch

            call laplac_init(laplac1,nsmatch1,npatch)
            call laplac_init(laplac2,nsmatch2,npatch)
    
        end subroutine row_laplac_init

end program Ultimatch_new_all
