!Module for the Wiener filters design using then Levinson recursion
!Note that in Shaping is assumed that the matrix is simmetric (i.e. forward = reverse(backward))
module Wiener_filters
contains
   subroutine Shaping(zauto,zpcross,filter,epsi)
	!Arguments of the subroutine
	real, dimension(:)		:: zauto,zpcross,filter
	real				:: epsi
	!Internal variables
	real, dimension(:),allocatable	:: forward_vector
	real				:: err, alfa, err_filt
	integer				:: i, j, k
	real :: debug

	allocate(forward_vector(size(zauto)))
	forward_vector(:) = 0.

	!Add the white noise to the main diagonal
	zauto(1) = (1 + epsi) * zauto(1)
	
	!Start the inversion
	!First iteration solves 1x1 problem
	filter(1) = zpcross(1) / zauto(1)
	forward_vector(1) = 1 / zauto(1)
	do i = 2, size(filter)
	   err = 0.
	   err_filt = 0.
	   do j = 2, i
	   	err = err + zauto(i-j+2) * forward_vector(j-1)
	   enddo
	   alfa = 1. / (1. - err**2.)
	   forward_vector(1:i) = alfa * forward_vector(1:i) - err * alfa * forward_vector(i:1:-1)
	   do k = 2, i
		err_filt = err_filt + zauto(i-k+2) * filter(k-1)
	   enddo
	   filter(1:i) = filter(1:i) + (zpcross(i) - err_filt) * forward_vector(i:1:-1)
	enddo
	


	deallocate(forward_vector)
   end subroutine Shaping
end module Wiener_filters
