!!$
!!$=head1 NAME
!!$
!!$nmo_mod - Perform nmo
!!$
!!$=head1 SYNOPSIS
!!$
!!$Initializer-C<call nmo_init(slowness,offset,o1,d1)>
!!$
!!$operator-C<call nmo_nearest(adj,tt,zz)>
!!$
!!$operator-C<call nmo_linear(adj,tt,zz)>
!!$
!!$=head1 INITIALIZER
!!$
!!$=head2 INPUT PARAMETERS
!!$
!!$=over 4
!!$
!!$=item slowness - C<real(:)>
!!$
!!$           Slowness(t)
!!$
!!$=item offset   - C<real(:)>
!!$
!!$           Offset
!!$
!!$=item o1       - real
!!$
!!$           First sample in time
!!$
!!$=item d1       - real
!!$
!!$           Sampling in time
!!$
!!$=item inverse  - logical
!!$
!!$           Wheter or not to do inverse
!!$
!!$=back
!!$
!!$=head1  OPERATORS
!!$
!!$=over 4
!!$
!!$=item adj   - logical
!!$
!!$        Whether or not do adjoint
!!$
!!$=item tt    - C<real(:,:)>
!!$
!!$        Input
!!$
!!$=item zz    - C<real(:,:)>
!!$
!!$        Output
!!$
!!$=back
!!$
!!$
!!$=head1 DESCRIPTION
!!$Nmo by nearest or linear interpolation neighbor for a gather
!!$
!!$=head1 COMMENTS
!!$
!!$NMO, its transpose, inverse, and inverse transpose by nearest neighbor.
!!$
!!$adj=.false. inv=.false.: zz(iz) = [NMO]                   tt(it)
!!$
!!$adj=.true. inv=.false.: tt(it) = [NMO adj]         zz(iz)
!!$
!!$adj=.false. inv=.true.:        tt(it) = [NMO inverse]           zz(iz)
!!$
!!$adj=.true. inv=.true.:        zz(iz) = [NMO inverse transpose] tt(it)
!!$
!!$
!!$=head1 SEE ALSO
!!$
!!$L<Nmo3d>
!!$
!!$=head1 LIBRARY
!!$
!!$B<sepvelanf90>
!!$
!!$=cut
module myhmo_mod 
	implicit none
	real,save,private :: slow
	real,save, pointer,private :: x(:) ! offset
	real,save, pointer,private :: twater(:) ! time to zero-offset water bottom
	real,save,private :: t0,dt
	integer,save,private :: nt,n2
	logical, save,private :: inv
	contains
	subroutine hmo_init (slowness, twb, offset, n1, o1, d1, inverse, first)
		real, intent (in)	:: offset(:)
		real, intent (in)	:: twb(:)
		real, intent (in)			:: slowness
		integer, intent (in)		:: n1
		real, intent (in)			:: o1, d1
		logical, intent (in)		:: inverse
		logical, intent (inout)		:: first

		inv=inverse
		t0=o1
		dt=d1
		nt=n1
		slow=slowness
		if (.not. first) then
			deallocate(x)
			deallocate(twater)
		end if
		allocate(x(size(offset)))
		allocate(twater(size(twb)))
		x=offset
		twater=twb
		first=.false.
		n2=size(x)
	end subroutine 
!>
!
	subroutine hmo_nearest (tt, zz)
		integer   it, iz, i2
		real   tt(:,:), zz(:,:)
		real :: tshift(n2)

		tshift = sqrt (twater * twater + x * x * slow * slow)

		do i2 = 1, n2
			do iz = 1, nt
				it = iz - (0.5 + tshift(i2) / dt) + (0.5 + twater (i2) / dt)
				if (it >= 1 .and. it <=nt ) then
					zz (it, i2) = tt (iz, i2)
				endif
			end do
		end do

	end subroutine 
end module 
