module slope

contains
subroutine slow(slowness,vert_slow_refrac,offset,wat_vel,tmax,d1,n1,depth)
  !Routine that gives the vertical slowness for each time sample 
  !This considers a constant velocity model and us the velocity picked during the filter design for the refracted arrival
   
  !Arguments of the routine
  real, dimension(:)		:: slowness
  real				:: vert_slow_refrac, offset, wat_vel, tmax, d1
  integer			:: n1, depth

  !Internal variables
  real(kind=8)			:: tmin_hyp, ray_par, time
  integer			:: tmin_hyp_ix, i


  !Gwdepth is in [dm] so must converted into [m]
  tmin_hyp = (offset + real(depth)/10.) / wat_vel -0.05! In this way I consider all the area under the direct arrival to be hyperbolic with the water velocity (time domain) t0 not = 0. 
  tmin_hyp_ix = nint(tmin_hyp / d1 + 1)
  !Set the local slope for refracted arrivals and the reflections with water velocity (NMO equation not HMO)
  if (tmin_hyp_ix <= n1) then
    slowness(1:tmin_hyp_ix-1) = 1. / vert_slow_refrac !Inverse of the vertical slowness
    time = tmin_hyp  ! NOTE the direct wave from source to receiver is not centered at t0=0.  
    do i = tmin_hyp_ix, n1
	ray_par = offset / (time * wat_vel**2.)
	time = time + d1
	slowness(i) = 1. / sqrt(wat_vel**-2. - ray_par**2.)
    enddo
    slowness(:) = slowness(:) * vert_slow_refrac!Renormalize because in the filter the slowness factor is already inside it.
  endif
end subroutine slow

end module slope
